/* main.c
 * Linux CAN-bus device driver.
 * Written by Arnaud Westenberg email:arnaud@wanadoo.nl
 * Rewritten for new CAN queues by Pavel Pisa - OCERA team member
 * email:pisa@cmp.felk.cvut.cz
 * This software is released under the GPL-License.
 * Version lincan-0.2  9 Jul 2003
 */

#define EXPORT_SYMTAB

#include <linux/autoconf.h>

#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/poll.h>
#include <linux/version.h>
#include <linux/autoconf.h>
#include <linux/interrupt.h>

#if (LINUX_VERSION_CODE <= KERNEL_VERSION(2,5,50))
#include <linux/wrapper.h>
#endif

#if !defined (__GENKSYMS__) 
#if (defined (MODVERSIONS) && !defined(NOVER))
#include <linux/modversions.h>
/*#include "../include/main.ver"*/
#endif
#endif

/*#undef CONFIG_DEVFS_FS*/

#ifdef CONFIG_DEVFS_FS
#include <linux/devfs_fs_kernel.h>
#include <linux/miscdevice.h>
#endif

#include "../include/can.h"
#include "../include/can_sysdep.h"
#include "../include/main.h"
#include "../include/modparms.h"
#include "../include/devcommon.h"
#include "../include/setup.h"
#include "../include/proc.h"
#include "../include/open.h"
#include "../include/close.h"
#include "../include/read.h"
#include "../include/select.h"
#include "../include/irq.h"
#include "../include/ioctl.h"
#include "../include/write.h"
#include "../include/finish.h"
#include "../include/fasync.h"

#define EXPORT_SYMTAB

/* Module parameters, some must be supplied at module loading time */
int major=CAN_MAJOR;
MODULE_PARM(major,"1i");
int minor[MAX_TOT_CHIPS]={-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
/*MODULE_PARM(minor, "1-" __MODULE_STRING(MAX_TOT_CHIPS)"i");*/
MODULE_PARM(minor, "1-" __MODULE_STRING(MAX_TOT_CHIPS_STR)"i");
int extended=0;
MODULE_PARM(extended,"1i");
int pelican=0;
MODULE_PARM(pelican,"1i");
int baudrate[MAX_TOT_CHIPS];
MODULE_PARM(baudrate, "1-" __MODULE_STRING(MAX_TOT_CHIPS_STR)"i");
char *hw[MAX_HW_CARDS]={NULL,};
MODULE_PARM(hw, "1-" __MODULE_STRING(MAX_HW_CARDS)"s");
int irq[MAX_IRQ]={-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
MODULE_PARM(irq, "1-" __MODULE_STRING(MAX_IRQ)"i");
unsigned long io[MAX_HW_CARDS]={-1,-1,-1,-1,-1,-1,-1,-1};
MODULE_PARM(io, "1-" __MODULE_STRING(MAX_HW_CARDS)"i");
int stdmask=0;
MODULE_PARM(stdmask, "1i");
int extmask=0;
MODULE_PARM(extmask, "1i");
int mo15mask=0;
MODULE_PARM(mo15mask, "1i");
int processlocal=0;
MODULE_PARM(processlocal, "1i");

/* Other module attributes */
#ifdef MODULE_SUPPORTED_DEVICE
MODULE_SUPPORTED_DEVICE("can");
#endif
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif
#ifdef MODULE_DESCRIPTION
MODULE_DESCRIPTION("Universal Linux CAN-bus device driver");
#endif

/* Global structures, used to describe the installed hardware. */
struct canhardware_t canhardware;
struct canhardware_t *hardware_p=&canhardware;
struct chip_t *chips_p[MAX_TOT_CHIPS];
struct msgobj_t *objects_p[MAX_TOT_MSGOBJS];
#ifdef CONFIG_DEVFS_FS
#if (LINUX_VERSION_CODE <= KERNEL_VERSION(2,5,60))
devfs_handle_t  devfs_handles[MAX_TOT_MSGOBJS];
#endif
#endif

/* Pointers to dynamically allocated memory are maintained in a linked list
 * to ease memory deallocation.
 */
struct mem_addr *mem_head=NULL;

struct file_operations can_fops=
{
 #if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0))
	owner:		THIS_MODULE,	
 #endif
	read:		can_read,
	write:		can_write,
	poll:		can_poll,
	ioctl:		can_ioctl,
	open:		can_open,
	release:	can_close,
  #ifdef CAN_ENABLE_KERN_FASYNC
	fasync:		can_fasync
  #endif /*CAN_ENABLE_KERN_FASYNC*/
};

EXPORT_SYMBOL(can_fops);

int init_module(void)
{
	int res=0,i=0, j;
	struct candevice_t *candev;
	struct chip_t *chip;

	if (parse_mod_parms())
		return -EINVAL;

	canqueue_kern_initialize();

	if (init_hw_struct())
		return -ENODEV;

	#ifdef CAN_DEBUG
		list_hw();
	#endif

	res=register_chrdev(major,DEVICE_NAME, &can_fops);
	if (res<0) {
		CANMSG("Error registering driver.\n");
		return -ENODEV;
	}

	for (i=0; i<hardware_p->nr_boards; i++) {
		candev=hardware_p->candevice[i];
		if (candev->hwspecops->request_io(candev)) 
			goto memory_error;
		candev->flags|=CANDEV_IO_RESERVED;
	}

	for (i=0; i<hardware_p->nr_boards; i++) {
		candev=hardware_p->candevice[i];
		if (candev->hwspecops->reset(candev)) 
			goto reset_error;
	}

	can_spin_lock_init(&hardware_p->rtr_lock);
	hardware_p->rtr_queue=NULL;

	for (i=0; i<hardware_p->nr_boards; i++) {
		candev=hardware_p->candevice[i];
		for(j=0; j<candev->nr_all_chips; j++) {
			if((chip=candev->chip[j])==NULL)
				continue;
			if(!chip->chipspecops->irq_handler)
				continue;
			
			if (request_irq(chip->chip_irq,chip->chipspecops->irq_handler,SA_SHIRQ,DEVICE_NAME,chip))
				goto interrupt_error;
			else {
				DEBUGMSG("Registered interrupt %d\n",chip->chip_irq);
				chip->flags |= CHIP_IRQ_SETUP;
			}
		}
		
		if (candev->flags & CANDEV_PROGRAMMABLE_IRQ)
			if (candev->hwspecops->program_irq(candev))
				goto interrupt_error;
	}

#ifdef CONFIG_PROC_FS
	if (can_init_procdir())
		goto proc_error;
#endif

#ifdef CONFIG_DEVFS_FS
        {
		char dev_name[32];
		int dev_minor;
        	for(i=0;i<MAX_TOT_MSGOBJS;i++) {
        		if(!objects_p[i]) continue;
           		dev_minor=objects_p[i]->minor;
		    #if (LINUX_VERSION_CODE <= KERNEL_VERSION(2,5,50))
			sprintf (dev_name, "can%d", dev_minor);
             		devfs_handles[i]=devfs_register(NULL, dev_name,
             			DEVFS_FL_DEFAULT, major, dev_minor,
		    		S_IFCHR | S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP,
		    		&can_fops, (void*)objects_p[i]);
		    #else
			devfs_mk_cdev(MKDEV(major, dev_minor), S_IFCHR | S_IRUGO | S_IWUGO, "can%d", dev_minor);
		    #endif
             	}
        }
#endif
	return 0;

#ifdef CONFIG_PROC_FS
	proc_error: ;
		CANMSG("Error registering /proc entry.\n");
		goto memory_error; 
#endif

	interrupt_error: ;
		CANMSG("Error registering interrupt line.\n");
		goto memory_error;

	reset_error: ;
		goto memory_error;

	memory_error: ;
		canhardware_done(hardware_p);

		res=unregister_chrdev(major,DEVICE_NAME);
		if (res<0)
			CANMSG("Error unloading CAN driver, error: %d\n",res);
		else
			CANMSG("No CAN devices or driver setup error.\n");
		return -ENODEV;

}

void cleanup_module(void)
{
	int res=0,i=0;

#ifdef CONFIG_PROC_FS
	if (can_delete_procdir())
		CANMSG("Error unregistering /proc/can entry.\n"); 
#endif

#ifdef CONFIG_DEVFS_FS
       	for(i=0;i<MAX_TOT_MSGOBJS;i++) {
	    #if (LINUX_VERSION_CODE <= KERNEL_VERSION(2,5,50))
       		if(devfs_handles[i])
       			devfs_unregister(devfs_handles[i]);
	    #else
		int dev_minor;
       		if(!objects_p[i]) continue;
       		dev_minor=objects_p[i]->minor;
		if(minor>=0)
			devfs_remove("can%d", dev_minor);
	    #endif
       	}
#endif

	canhardware_done(hardware_p);

	if ( can_del_mem_list() ) 
		CANMSG("Error deallocating memory\n");

	res=unregister_chrdev(major,DEVICE_NAME);
	if (res<0)
		CANMSG("Error unregistering CAN driver, error: %d\n",res);
}
