/* setup.c
 * Linux CAN-bus device driver.
 * Written by Arnaud Westenberg email:arnaud@wanadoo.nl
 * Rewritten for new CAN queues by Pavel Pisa - OCERA team member
 * email:pisa@cmp.felk.cvut.cz
 * This software is released under the GPL-License.
 * Version lincan-0.2  9 Jul 2003
 */ 

#include "../include/can.h"
#include "../include/can_sysdep.h"
#include "../include/main.h"
#include "../include/devcommon.h"
#include "../include/setup.h"
#include "../include/finish.h"

extern int sja1000_register(struct chipspecops_t *chipspecops);
extern int sja1000p_register(struct chipspecops_t *chipspecops);
extern int i82527_register(struct chipspecops_t *chipspecops);

int init_device_struct(int card);
int init_hwspecops(struct candevice_t *candev);
int init_chip_struct(struct candevice_t *candev);
int init_obj_struct(struct candevice_t *candev, struct chip_t *hostchip, int minorbase);
int init_chipspecops(struct candevice_t *candev, int chipnr);

void *can_checked_malloc(size_t size)
{
	struct mem_addr *mem_new;
	void *address_p;
	
	address_p=kmalloc(size,GFP_KERNEL);
	if(address_p == NULL) {
		CANMSG("can_checked_malloc: out of the memory\n");
		return NULL;
	}

#ifdef DEBUG_MEM
	DEBUGMSG("can_checked_malloc: allocated %d bytes at %p, mem_head=%p\n",
			(int)size, address_p, mem_head);
#endif

	mem_new=(struct mem_addr *)kmalloc(sizeof(struct mem_addr),GFP_KERNEL);
	if (mem_new == NULL) {
		CANMSG("can_checked_malloc: memory list allocation error.\n");
		kfree(address_p);
		return NULL;
	}
	mem_new->next=mem_head;
	mem_new->address=address_p;
	mem_new->size=size;
	mem_head=mem_new;

	return address_p;
}

int can_checked_free(void *address_p)
{
	struct mem_addr **mem_pptr;
	struct mem_addr *mem_del=NULL;

#ifdef DEBUG_MEM
	DEBUGMSG("can_checked_free %p, mem_head=%p\n", address_p, mem_head);
#endif

	for(mem_pptr = &mem_head; (mem_del = *mem_pptr); mem_pptr = &mem_del->next) {
		if (mem_del->address != address_p)
			continue;
		*mem_pptr=mem_del->next;
		kfree(mem_del);
		kfree(address_p);
		return 0;
	}
	
	CANMSG("can_checked_free: address %p not found on the mem list\n", address_p);
	
	kfree(address_p);
	return -1;
}


int can_del_mem_list(void)
{
	struct mem_addr *mem;

#ifdef DEBUG_MEM
	DEBUGMSG("can_del_mem_list, mem_head=%p\n", mem_head);
#endif
	if(mem_head == NULL) {
		CANMSG("can_del_mem_list: no entries on the list - OK\n");
		return 0;
	}

	while((mem=mem_head) != NULL) {
		mem_head=mem->next;
		CANMSG("can_del_mem_list: deleting %p with size %d\n",
			mem->address, (int)mem->size);
		kfree(mem->address);
		kfree(mem);
	}
	
	return 0;
}

int can_request_io_region(unsigned long start, unsigned long n, const char *name)
{
    #if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0))
	if(check_region(start,n)) return 0;
	request_region(start,n,name);
	return 1;
    #else
	return (request_region(start,n,name))?1:0;
    #endif
}

void can_release_io_region(unsigned long start, unsigned long n)
{
	release_region(start,n);
}

int can_request_mem_region(unsigned long start, unsigned long n, const char *name)
{
    #if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0))
	return 1;
    #else
	return (request_mem_region(start,n,name))?1:0;
    #endif
}

void can_release_mem_region(unsigned long start, unsigned long n)
{
    #if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0))
	return;
    #else
	release_mem_region(start,n);
    #endif
}

/* This function shifts all base address structures acording to address
   translation between physical and virtual address mappings */
int can_base_addr_fixup(struct candevice_t *candev, unsigned long new_base)
{
	unsigned long offs;
	int i, j;
	
	offs=new_base-candev->dev_base_addr;
	candev->dev_base_addr=new_base;
	for(i=0;i<candev->nr_all_chips;i++){
		candev->chip[i]->chip_base_addr += offs;
		for(j=0;j<candev->chip[i]->max_objects;j++)
			candev->chip[i]->msgobj[j]->obj_base_addr += offs;
	}
	return 0;
}

/* The function init_hw_struct is used to initialize the hardware structure. */
int init_hw_struct(void)
{
	int i=0;

	hardware_p->nr_boards=0;
	while ( (hw[i] != NULL) & (i < MAX_HW_CARDS) ) {
		hardware_p->nr_boards++;

		if (init_device_struct(i)) {
			CANMSG("Error initializing candevice_t structures.\n");
			return -ENODEV;
		}
		i++;
	}

	return 0;
}

/* The function init_device_struct is used to initialize a single device 
 * structure.
 */
int init_device_struct(int card)
{
	struct candevice_t *candev;
	int ret;
	
	candev=(struct candevice_t *)can_checked_malloc(sizeof(struct candevice_t));
	if (candev==NULL)
		return -ENOMEM;

        memset(candev, 0, sizeof(struct candevice_t));

	hardware_p->candevice[card]=candev;
	candev->candev_idx=card;

	candev=candev;

	candev->hwname=hw[card];
	candev->io_addr=io[card];
	candev->dev_base_addr=io[card];

	candev->hwspecops=(struct hwspecops_t *)can_checked_malloc(sizeof(struct hwspecops_t));
	if (candev->hwspecops==NULL)
		goto error_nomem;

	memset(candev->hwspecops, 0, sizeof(struct hwspecops_t));

	if (init_hwspecops(candev))
		goto error_nodev;

	if (candev->hwspecops->init_hw_data(candev))
		goto error_nodev;

	if (init_chip_struct(candev))
		goto error_nodev;

	return 0;

    error_nodev:
	candevice_done(candev);
	ret=-ENODEV;
	goto error_both;

    error_nomem:
	ret=-ENOMEM;

    error_both:
	hardware_p->candevice[card]=NULL;
	can_checked_free(candev);
	return ret;
    	
}

/* The function init_chip_struct is used to initialize all chip_t structures
 * on one hardware board.
 */
int init_chip_struct(struct candevice_t *candev)
{
	static int irq_count=0;
	int i=0;

	/* Alocate and initialize the chip structures */
	for (i=0; i < candev->nr_all_chips; i++) {
		candev->chip[i]=(struct chip_t *)can_checked_malloc(sizeof(struct chip_t));
		if (candev->chip[i]==NULL)
			return -ENOMEM;

        	memset(candev->chip[i], 0, sizeof(struct chip_t));
		
		candev->chip[i]->write_register=candev->hwspecops->write_register;
		candev->chip[i]->read_register=candev->hwspecops->read_register;

		candev->chip[i]->chipspecops=can_checked_malloc(sizeof(struct chipspecops_t));
		if (candev->chip[i]->chipspecops==NULL)
			return -ENOMEM;

                chips_p[irq_count]=candev->chip[i];
		candev->chip[i]->chip_idx=i;
		candev->chip[i]->hostdevice=candev;
		candev->chip[i]->chip_irq=irq[irq_count];
		candev->chip[i]->baudrate=baudrate[irq_count]*1000;
		if(!candev->chip[i]->baudrate)
			candev->chip[i]->baudrate=baudrate[0]*1000;
		candev->chip[i]->flags=0x0;

		candev->hwspecops->init_chip_data(candev,i);

		if (init_chipspecops(candev,i))
			return -ENODEV;

		init_obj_struct(candev, candev->chip[i], minor[irq_count]);

		irq_count++;
	} 

	return 0;
}

int init_obj_struct(struct candevice_t *candev, struct chip_t *hostchip, int minorbase)
{
	struct canque_ends_t *qends;
	static int obj_count=0;
	int i,max_objects;
	struct msgobj_t *obj;

	max_objects=hostchip->max_objects;
	for (i=0; i<max_objects; i++) {
		obj=(struct msgobj_t *)can_checked_malloc(sizeof(struct msgobj_t));
		hostchip->msgobj[i]=obj;
		if (obj == NULL) 
			return -ENOMEM;
			
        	memset(obj, 0, sizeof(struct msgobj_t));

		atomic_set(&obj->obj_used,0);
		INIT_LIST_HEAD(&obj->obj_users);
		init_timer(&obj->tx_timeout);
		
		qends = (struct canque_ends_t *)can_checked_malloc(sizeof(struct canque_ends_t));
		if(qends == NULL) return -ENOMEM;
		memset(qends, 0, sizeof(struct canque_ends_t));
		obj->hostchip=hostchip;
		obj->object=i+1;
		obj->qends=qends;
		obj->tx_qedge=NULL;
		obj->tx_slot=NULL;
		obj->flags = 0x0;

		canqueue_ends_init_chip(qends, hostchip, obj);
		
		if (minorbase == -1) minorbase=obj_count;
		if ((minorbase >= 0) && (minorbase+i<MAX_TOT_MSGOBJS)){
		  objects_p[minorbase+i]=obj;
		  obj->minor=minorbase+i;
		} else obj->minor=-1;

		candev->hwspecops->init_obj_data(hostchip,i);

		obj_count++;
	}
	return 0;
}


int init_hwspecops(struct candevice_t *candev)
{
	const struct boardtype_t *brp;
	
	brp = boardtype_find(candev->hwname);
	
	if(!brp) {
		CANMSG("Sorry, hardware \"%s\" is currently not supported.\n",candev->hwname);
		return -EINVAL;
	}
	
	brp->board_register(candev->hwspecops);

	return 0;
}

int init_chipspecops(struct candevice_t *candev, int chipnr)
{
	if (!strcmp(candev->chip[chipnr]->chip_type,"i82527")) {
		candev->chip[chipnr]->max_objects=15;
		i82527_register(candev->chip[chipnr]->chipspecops);
	} 
	if (!strcmp(candev->chip[chipnr]->chip_type,"sja1000")) {
		candev->chip[chipnr]->max_objects=1;
		sja1000_register(candev->chip[chipnr]->chipspecops);
	}
	if (!strcmp(candev->chip[chipnr]->chip_type,"sja1000p")) {
		candev->chip[chipnr]->max_objects=1;
		sja1000p_register(candev->chip[chipnr]->chipspecops);
	}

	return 0;
}
